#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Linked List
		Realisierung eines Datentyps mit den Operationen von den Folien.
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 03-Sortierung durch Verteilen
	Ca. Folie(n): 54
"""

# Datentypen
mutable struct ListElement 
	first :: Any
	rest :: Union{ListElement, Nothing}
end

mutable struct LinkedList 
	internal_repr :: Union{ListElement, Nothing}
	length :: Int
end

# Funktionen
function make_list()
	return LinkedList(nothing, 0)
end

function insert(e, l)
	l.internal_repr = ListElement(e, l.internal_repr)
	l.length += 1
end

function delete(e, l)
	c = l.internal_repr 
	last = nothing
	while !isnothing(c)
		if c.first == e 
			if isnothing(last)
				l.internal_repr = c.rest 
			else
				last.rest = c.rest
			end
			l.length -= 1
			break
		end
		last = c
		c = c.rest
	end
end

function first(l)
	if isnothing(l.internal_repr)
		throw(BoundsError(l, 1))
	else 
		return l.internal_repr.first
	end
end

function delete_first(l)
	if isnothing(l.internal_repr)
		throw(BoundsError(l, 1))
	else
		l.internal_repr = l.internal_repr.rest
		l.length -= 1
	end
end

function length(l)
	return l.length
end

function empty_list(l)
	return l.length == 0
end

# Iteration mit Julia ermoeglichen
function Base.iterate(l::LinkedList)
	if isnothing(l.internal_repr)
		return nothing 
	else 
		return l.internal_repr.first, l.internal_repr.rest
	end 
end 
function Base.iterate(l::LinkedList, state::ListElement)
	return state.first, state.rest 
end
function Base.iterate(l::LinkedList, state::Nothing)
	return nothing
end 
function Base.length(l::LinkedList)
	return l.length
end

# Beispielaufrufe
# 	Liste erstellen und etwas einfuegen
my_list = make_list()
insert("O", my_list)
insert("L", my_list)
insert("L", my_list)
insert("A", my_list)
insert("H", my_list)
# 	ausgeben
println(my_list)
#	Iteration nutzen
println(reduce(*, my_list))
#	Elemente loeschen und Groesse beobachten
delete("A", my_list)
println(first(my_list), ' ', length(my_list))
delete_first(my_list)
println(first(my_list), ' ', length(my_list), ' ', empty_list(my_list))
#	Iteration nutzen
for e in my_list
	print(e)
end
println()

"""
	Operation cons 
	Foliensatz: 03-Sortierung durch Verteilen
	Ca. Folie(n): 58
"""
function cons(e, l)
	return LinkedList(
		ListElement(e, l.internal_repr),
		l.length+1
	)
end

my_other_list = cons("A", my_list)

println(map(*, my_list))
println(map(*, my_other_list))