#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Set mit eigener Keyfunktion
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 06-Mengen
	Ca. ab Folie(n): 6
"""

struct KeySet 
	internal_repr :: Vector{Any}
	key :: Function

	KeySet() = KeySet([])
	KeySet(elements) = KeySet(elements, identity)
	function KeySet(elements, key::Function)
		s = new([], key)
		for e in elements
			insert(e, s)
		end
		return s
	end
end 

function key(set::KeySet)
	return set.key
end

# Iteratoren fuer KeySet definieren
function Base.iterate(s::KeySet) 
	if isempty(s.internal_repr)
		return nothing 
	else 
		return s.internal_repr[1], 2
	end 
end 
function Base.iterate(s::KeySet, index::Int) 
	if index <= length(s)
		return s.internal_repr[index], index+1
	else 
		return nothing
	end
end
function Base.length(s::KeySet) 
	return length(s.internal_repr)
end

function Base.show(io::IO, s::KeySet) 
	println(io, "{ " * fold( (i, e) -> i * string(", ", e), s, "")[3:end] * " }")
end

# Operationen auf den Mengen 
function test(k, s::KeySet)
	# Dieses test() ist allgemein auch fuer anderen Mengen 
	# anwendbar, denn es nutzt nur die Funktion search().
	#	Aus Gruenden der Performance koennen anderen 
	#	Mengen test() natuerlich ueberschreiben.
	e = search(k, s)
	return !isnothing(e) && key(s)(e) == k 
end

function search(k, s::KeySet)
	for e in s
		if key(s)(e) == k 
			return e
		end 
	end
	return nothing
end

function delete(k, s::KeySet)
	for (i, e) in enumerate(s)
		if key(s)(e) == k 
			deleteat!(s.internal_repr, i)
		end 
	end
end

function insert(e, s::KeySet)
	k = key(s)(e)
	if test(k, s)
		delete(k, s)
	end
	push!(s.internal_repr, e)
end

function map(f::Function, s::KeySet)
	for i in length(s.internal_repr)
		s.internal_repr[i] = f(s.internal_repr[i])
	end
end

function fold(f::Function, s::KeySet, init)
	r = init 
	for e in s
		r = f(r, e)
	end
	return r
end

# Beispielaufrufe
s = KeySet([1,2,3,2,4])
print(s)
println("test(5) = ", test(5, s), ", test(2) = ",  test(2, s), ", search(2) = ", search(2, s))
insert(12, s)
insert(3, s)
delete(2, s)
print(s)
map((x)->x^2, s)
print(s)
function max(e, e_)
	return if e > e_ e else e_ end
end
println("max = ", fold(max, s, 0))

println()
s = KeySet([10, 15, 20, 25, 30], (x) -> x ÷ 10)
print(s)