#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Vergleiche zwischen Objekten
	Julia Iteratoren

	Foliensatz: 06-Mengen
	Ca. ab Folie(n): 9
"""

"""
Vergleiche zwischen Objekten 
	=> Falls nicht anders angegeben, wird haeufig der Pointer verglichen
		Manchmal ist Julia auch schlau und vergleicht die Elemente so, wie man
		es vermuten wuerde, aber darauf sollte man sich nicht verlassen!
	=> Kann man in Julia direkt an den Typ des Objekten binden

	=> Fuer die meisten eigenen Typen von Julia *sinnvoll* definiert
		z.B. gilt "ab" == "ab" oder [1,2,3] == [1,2,3] 
"""

println("[1,2,3] == [1,2,3]: ", [1,2,3] == [1,2,3])
println("\"ab\" == \"ab\": ", "ab" == "ab")
println("Set([1,2]) == Set([2,1]): ", Set([1,2]) == Set([2,1]))
println()

# Eigener Typ mit selbstdefinierten Vergleichsoperationen
struct MyType
	values :: Array{Any}
end
function Base.:(==)(o::MyType, o_::MyType)
	return o.values == o_.values # Array werden hier elementweise verglichen
end
function Base.:(<=)(o::MyType, o_::MyType)
	return sum(o.values) <= sum(o_.values)
end
function Base.:(<)(o::MyType, o_::MyType)
	return sum(o.values) < sum(o_.values)
end
function Base.:(>)(o::MyType, o_::MyType)
	return sum(o.values) > sum(o_.values)
end
function Base.:(>=)(o::MyType, o_::MyType)
	return sum(o.values) >= sum(o_.values)
end

# Eigener Typ ohne selbstdefinierte Vergleichsoperationen  
struct MyOtherType
	values :: Array{Any}
end

println("MyType([12]) == MyType([12]): ", MyType([12]) == MyType([12]))
println("MyOtherType(12) == MyOtherType(12): ",
	MyOtherType([12]) == MyOtherType([12])
)
println("MyType([6, 6]) >= MyType([12]): ", MyType([6, 6]) >= MyType([12]))
try
	println("MyOtherType([6, 6]) >= MyOtherType([12]): ", 
		MyOtherType([6, 6]) >= MyOtherType([12])
	)
catch e
	println("Achtung, MyOtherType([6, 6]) >= MyOtherType([12]) geht nicht!")
	println("\t", typeof(e))
end
println("MyType([6, 6]) > MyType([12]): ", MyType([6, 6]) > MyType([12]))
println()
println()

"""
Iteration ueber Objekte
	=> Definieren der Funktion Base.iterate() fuer den Typ
	=> Erlaubt dann for-Syntax in Julia und auch Nutzung mit Funktionen wie map()

	=> Sinnvoll auch die Funktion Base.length() mit zu definieren!
"""

struct MyIteratableType 
	elements :: Array{Any}
end

# Iteratoren definieren
function Base.iterate(iter::MyIteratableType) 
	if isempty(iter.elements)
		return nothing 
	else 
		return iter.elements[1], 2
	end 
end 
function Base.iterate(iter::MyIteratableType, index::Int) 
	if index <= length(iter)
		return iter.elements[index], index+1
	else 
		return nothing
	end
end
function Base.length(iter::MyIteratableType) 
	return length(iter.elements)
end

# Iteration nutzen
mit = MyIteratableType([1,2,3,4,5,6])

# 	For-Schleifen:
for element in mit
	print(element)
end
println()
for element ∈ mit
	print(element)
end
println()
# 	.. werden intern zu:
#		(let erzeugt einen Skopus, sodass element_state danach nicht mehr 
#		vefuegbar ist)
let element_state = iterate(mit)
	while !isnothing(element_state)
		(element, state) = element_state

		# Inhalt der Schleife
		print(element)

		element_state = iterate(mit, state)
	end
end
println()

# 	- `return nothing` beendet Iteration 
#	- `return element, state` gibt `element` an das Innere der Schleife
#	- `state` wird genutzt um weiter forwaerts iterieren zu koennen

# Funktionen die Iteration nutzen
println(map((x) -> x^2, mit))
println("reduce(+, mit) = ", reduce(+, mit))
