#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Selbstanordnende Listen

		Fuer eine kuerzere Implementierung nehmen wir hier an, dass 
		die Identitaet als Schluessel genutzt wird.
		Andernfalls muss man den Schluessel durch Julia auf Typebene der Elemente 
		definieren, z.B. `function Base.:(==)(e::Elem, e_::Elem) key(e) == key(e_) end`.

		Weiterhin nutzen wir keine Liste, sondern die Julia-internen Vektoren als Basisdatenstruktur.

	Foliensatz: 06-Mengen
	Ca. Folie(n): 20
"""

const MF_RULE = 1
const T_RULE = 2
const FC_RULE = 3

struct SelfOrderingList 
	internal_repr :: Vector{Any}
	rule :: Int
	counts :: Union{Nothing, Vector{Int}}

	function SelfOrderingList(elements, rule, counts)
		if !(rule in (MF_RULE, T_RULE, FC_RULE))
			throw(ArgumentError("Unknown rule given!"))
		end
		l = new([], rule, counts)
		for e in elements
			insert(e, l)
		end
		return l
	end
end 


function MFList(elements)
	return SelfOrderingList(elements, MF_RULE, nothing)
end
function TList(elements)
	return SelfOrderingList(elements, T_RULE, nothing)
end
function FCList(elements)
	return SelfOrderingList(elements, FC_RULE, Int[])
end

# Iteratoren definieren
function Base.iterate(s::SelfOrderingList) 
	if isempty(s.internal_repr)
		return nothing 
	else 
		return s.internal_repr[1], 2
	end 
end 
function Base.iterate(s::SelfOrderingList, index::Int) 
	if index <= length(s)
		return s.internal_repr[index], index+1
	else 
		return nothing
	end
end
function Base.length(s::SelfOrderingList) 
	return length(s.internal_repr)
end

function Base.show(io::IO, l::SelfOrderingList) 
	if l.rule == MF_RULE 
		name = "Move to front:   "
	elseif l.rule == T_RULE
		name = "Transpose:       "
	else 
		name = "Frequency count: "
	end
	str = ""
	for (i, e) in enumerate(l)
		str *= string(", ", e)
		str *= if l.rule == FC_RULE string(" (", l.counts[i], ")") else "    " end
	end
	println(io, name, "{ ", str[3:end], " }")
end

# Operationen 
function search(e, l::SelfOrderingList, move_elements=true)
	for (i, e_) in enumerate(l)
		if e_ == e 
			if move_elements # Regeln anwenden (ausser beim ersten einfuegen am Anfang)!
				if l.rule == MF_RULE 
					j = i
					while j > 1 # bis nach vorne tauschen
						l.internal_repr[j], l.internal_repr[j-1] = l.internal_repr[j-1], l.internal_repr[j]
						j -= 1
					end
				elseif l.rule == T_RULE && i > 1 # sofer nicht schon vorne, einmal tauschen
					l.internal_repr[i], l.internal_repr[i-1] = l.internal_repr[i-1], l.internal_repr[i]
				elseif l.rule == FC_RULE
					l.counts[i] += 1 # count anpassen
					j = i
					while j > 1 && l.counts[j] >= l.counts[j-1] # bis counts wieder absteigend
						l.internal_repr[j], l.internal_repr[j-1] = l.internal_repr[j-1], l.internal_repr[j]
						l.counts[j], l.counts[j-1] = l.counts[j-1], l.counts[j]
						j -= 1
					end
				end
			end
			return e
		end 
	end
	return nothing
end

function delete(e, l::SelfOrderingList)
	for (i, e_) in enumerate(l)
		if e_ == e 
			deleteat!(l.internal_repr, i)
			if l.rule == FC_RULE
				deleteat!(l.counts, i)
			end
		end 
	end
end

function insert(e, l::SelfOrderingList)
	if !isnothing(search(e, l, false))
		delete(e, l)
	end
	push!(l.internal_repr, e)
	if l.rule == FC_RULE
		push!(l.counts, 0)
	end
end


# Beispielaufrufe
elems = [1,2,3,4,3,5]
mf = MFList(elems)
t = TList(elems)
fc = FCList(elems)
println(mf, t, fc)

println("search 3:")
search(3, mf)
search(3, t)
search(3, fc)
println(mf, t, fc)

println("search 3:")
search(3, mf)
search(3, t)
search(3, fc)
println(mf, t, fc)

println("search 2:")
search(2, mf)
search(2, t)
search(2, fc)
println(mf, t, fc)