#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Einfaches Hashing (Zuordung) ohne Behandlung von Kollisionen
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Datentypen nutzen!

	Foliensatz: 09-Assoziation-von-Objekten
	Folie(n): 9-11
"""

include("./../Utils/generator.jl")

struct Dictionary
	internal_repr :: Array{Any}
	hash :: Function
end

function initialize_dictionary(
		init_values::Vector{<:Pair},
		hash_function::Function=identity,
		length::Int=17
	) :: Dictionary

	inner_hash_function = function (k::Int)
		return (hash_function(k) % length) + 1
	end

	d = Dictionary(
		Array{Any}(nothing, length),
		inner_hash_function
	)

	for (k,e) in init_values
		insert(k, e, d)
	end

	return d
end

function insert(k::Int, e::Any, d::Dictionary) :: Nothing
	T = d.internal_repr
	T[d.hash(k)] = (k,e)

	nothing
end

function delete(k::Int, d::Dictionary) :: Nothing
	T = d.internal_repr
	T[d.hash(k)] = nothing

	nothing
end

function lookup(k::Int, d::Dictionary) :: Any
	T = d.internal_repr
	t = T[d.hash(k)]
	if isnothing(t)
		return nothing
	else 
		return t[2]
	end
end

function Base.keys(d::Dictionary)
	return Generator( 
		(t) -> t[1], # Nimm jeweils das erste Element (key) aus den Tupeln
		d.internal_repr, # Array mit den Tupeln
		(t) -> !isnothing(t) # Nimm nur Elemente, die nicht nothing sind
	)
end

function Base.values(d::Dictionary)
	return Generator(
		(t) -> t[2], # Nimm jeweils das zweite Element (value) aus den Tupeln
		d.internal_repr, # Array mit den Tupeln
		(t) -> !isnothing(t) # Nimm nur Elemente, die nicht nothing sind
	)
end

function Base.pairs(d::Dictionary)
	return Generator(
		(t) -> t[1] => t[2], # Mache Pairs aus den Tupeln
		d.internal_repr, # Array mit den Tupeln
		(t) -> !isnothing(t) # Nimm nur Elemente, die nicht nothing sind
	)
end

function Base.show(io::IO, d::Dictionary)
	str = "\tHash/Index : Key => Value\n"
	for (hash, value) in enumerate(d.internal_repr)
		if !isnothing(value)
			str *= "\t$hash : $(value[1]) => $(value[2]),\n"
		end
	end
	print(io, "Dictionary:\n", str) 
end

# Beispielaufrufe
dict = initialize_dictionary([1 => "one", 2 => "two", 30 => "thirty"])
println(dict)

println("insert(17, \"seventeen\", dict)")
insert(17, "seventeen", dict)
println(dict)

println("insert(18, \"eighteen\", dict)")
insert(18, "eighteen", dict)
println(dict)

println( "lookup(30, dict) : ", lookup(30, dict))
println( "lookup(1, dict) : ", lookup(1, dict))
println()

println("delete(17, dict)")
delete(17, dict)
println(dict)

# Iteration
println("values(dict):")
for v in values(dict)
	print(v, ", ")
end
println("\n")

println("keys(dict):")
for k in keys(dict)
	print(k, ", ")
end
println("\n")

println("pairs(dict):")
for (k, v) in pairs(dict)
	print(k, " => ", v, ", ")
end
println("\n")

