
#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Beispiele fuer die hier genutze Graphdatenstruktur
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")

# initialize graph
a = Graph( ['A' => "B", "B" => "C", "C" => 'A'], "a-Graph" )
b = DirectedGraph( ["A" => "B", "B" => "C", "C" => "A"], "a-DirectedGraph" )
c = WeightedGraph( ["A" => "B", "B" => "C", "C" => "A"], [1,2,3], "a-WeightedGraph" )
d = DirectedWeightedGraph(["A" => "B", "B" => "C", "C" => "A"], [1,2,3], "a-DirectedWeightedGraph" )
print(a, b, c, d)

# add nodes 
insert("D", a)
insert("D", b)
insert("D", c)
insert("D", d)
#	add edges
insert("C", "D", a)
insert("C" => "D", b)
insert("C", "D", c, 12)
insert("C" => "D", d, 12)
print(a, b, c, d)
println()

# search and test nodes 
println("\"A\" in a == ", "A" in a)
println("\'A\' in a == ", 'A' in a)
println("\'A\' in b == ", 'A' in b)
println("find(\"A\", a) == ", find("A", a))
println("a[\"A\"] == ", a["A"])
# find('A', a) geht nicht, da wir nach den Namen suchen ('A' ist ja ein Objekt!)
println()

# search and test edges
println("find(\"A\", \"B\", a) == ", find("A", "B", a), " == find(\"B\", \"A\", a) == ", find("B", "A", a))
println("find('A', \"B\", a) == ", find('A', "B", a) ," == find(\"B\", 'A', a) == ", find("B", 'A', a))
println("find(\"A\", \"B\", b) == ", find("A", "B", b), " != find(\"B\", \"A\", b) == ", find("B", "A", b))
println("find(\"A\" => \"B\", a) == ", find("A" => "B", a))
println("a[\"A\", \"B\"] == ", a["A", "B"])
println("c[\"B\" => \"C\"] == ", c["B" => "C"])
println("a[\"D\", \"B\"] == ", a["D", "B"])
println()

# iterate nodes
print("nodes(a)\n\t")
for n in nodes(a)
	print("$n, $(typeof(n)); ")
end 
print("\nnodes(a, true)\n\t")
for n in nodes(a, true)
	print("$n, $(typeof(n)); ")
end 
println()
println()

# iterate all edges 
println("edges(a)")
for (u,v) in edges(a)
	println("\tu=$u, $(typeof(u)); v=$v, $(typeof(v))")
end 
println("edges(c, true)")
for (u,v,w) in edges(c, true)
	println("\tu=$u, $(typeof(u)); v=$v, $(typeof(v)), weight=$w")
end 
println()

println("edges_out('A', a, true)")
for (u,v) in edges_out('A', a)
	println("\tu=$u, $(typeof(u)); v=$v, $(typeof(v))")
end 
println("edges_out('A', a, true)")
for (u,v) in edges_out('A', a, true)
	println("\tu=$u, $(typeof(u)); v=$v, $(typeof(v))")
end 
println("edges_in(\"B\", b)")
for (u,v) in edges_in("B", b)
	println("\tu=$u, v=$v")
end 
println("edges_out(\"B\", b)")
for (u,v) in edges_out("B", b)
	println("\tu=$u, v=$v")
end 
println("edges_in(\"B\", d)")
for (u,v,w) in edges_in("B", d)
	println("\tu=$u, v=$v, w=$w")
end 
println("edges_out(\"B\", d)")
for (u,v,w) in edges_out("B", d)
	println("\tu=$u, v=$v, w=$w")
end 

# remove nodes
remove('A', a)
remove("A", b)
#	remove edges
remove("B", "C", c)
remove("B" => "C", d)
print(a, b, c, d)