#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Breitensuche
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 29
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Queue analog zu Foliensatz 03 laden
include("./../Utils/queue.jl")

mutable struct BfsNode
	d::Float32
	parent::Union{Nothing, BfsNode}
	value::Any
	BfsNode(value) = new(Inf, nothing, value)
end
function string(v::BfsNode)
	return v.value
end

function bfs(s::BfsNode, g::Graph)
	s.d = 0 # s hat Distanz 0 zu sich
	s.parent = s # s ist sein eigener Vater
	q = make_queue([s]) # Queue zu besuchender Knoten
	while !empty_queue(q) # solange q nicht leer
		u = dequeue(q) # nimm Knoten nach FIFO Regel
		for (u, v) in edges_out(u, g)
			if isnothing(v.parent) # v schon besucht?
				enqueue(v, q) # nein, dann in q hinten einfuegen
				v.d = u.d + 1
				v.parent = u
			end
		end
	end
end

function bfs_print(g::Graph)
	println(g)
	println("\tBFS values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\td[$(u.value)] = $(u.d), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end

# Beispielaufrufe
a, b, c = BfsNode("a"), BfsNode("b"), BfsNode("c")
g = DirectedGraph([b => c, c => a], "b-BfsGraph-i")
bfs(b, g)
bfs_print(g)

d = BfsNode("d")
g_ = DirectedGraph([b => c, b => d, c => a], "b-BfsGraph-ii")
bfs(b, g_)
bfs_print(g_)