#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	=> Tiefensuche
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 30, 41

	=> Pruefung ob ein Graph ein DAG
	
	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 50
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")

mutable struct DfsCount 
	dfsPos :: Int
	finishingTime :: Int 

	DfsCount() = new(1,1)
end
mutable struct DfsNode
	value::Any
	marked::Bool
	dfsNum::Int
	finishTime::Int

	DfsNode(value) = new(value, false, 0, 0)
end
function string(v::DfsNode)
	return v.value
end

function dfs(s::DfsNode, g::Graph)
	for v in nodes(g)
		v.marked = false
	end
	# init()
	dfsCount = DfsCount()

	s.marked = true
	# root(s)
	s.dfsNum = dfsCount.dfsPos
	dfsCount.dfsPos += 1

	dfs(s, s, g, dfsCount)
end
function dfs(u::DfsNode, v::DfsNode, g::Graph, dfsCount::DfsCount)
	for (_, w) in edges_out(v, g)
		if w.marked  
			# handle_non_tree_edge(v, w)
		else
			# traverse_tree_dge(v,w)
			w.dfsNum = dfsCount.dfsPos
			dfsCount.dfsPos += 1

			w.marked = true 
			dfs(v, w, g,  dfsCount)
		end
	end
	# backtrack(u, v)
	v.finishTime = dfsCount.finishingTime
	dfsCount.finishingTime += 1
end

function dfs_print(g::Graph)
	println(g)
	println("\tDFS values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		println("\t\tdfsNum[$(u.value)] = $(u.dfsNum), finishTime[$(u.value)] = $(u.finishTime) ")
	end
end

# Beispielaufrufe
A, B, C = DfsNode("A"), DfsNode("B"), DfsNode("C")
graph_i = DirectedGraph([B => C, C => A], "c-DfsGraph-i")
dfs(B, graph_i)
dfs_print(graph_i)

# Graph anlog zu Folie 41
a, b, c, d, e, f = DfsNode("a"), DfsNode("b"), DfsNode("c"), DfsNode("d"), DfsNode("e"), DfsNode("f")
g, h, i, j, k = DfsNode("g"), DfsNode("h"), DfsNode("i"), DfsNode("j"), DfsNode("k")
graph_ii = DirectedGraph([
	a => b, a => c, b => d, c => i, c => f, d => g, d => e,
	e => h, e => f, e => b, f => i, g => h, h => j, h => k,
	i => h, k => i
], "c-DfsGraph-ii")
dfs(a, graph_ii)
dfs_print(graph_ii)

# Pruefung ob ein Graph ein DAG
function isDag(g::Graph)
	for v in nodes(g)
		v.marked = false
	end
	# init()
	dfsCount = DfsCount()

	# stelle sicher, dass alle Knoten besucht werden
	for s in nodes(g)
		if !s.marked 
			s.marked = true
			# root(s)
			s.dfsNum = dfsCount.dfsPos
			dfsCount.dfsPos += 1

			dfs(s, s, g, dfsCount)
		end 
	end

	# pruefe auf Rueckwaertskante
	for (v, w) in edges(g)
		if !(v.finishTime > w.finishTime)
			return false # eine Rueckwaertskante gefunden, kein DAG!
		end
	end
	return true # ist ein DAG, denn fuer alle Kanten: finishTime[v] > finishTime[w]
end

# Beispielaufrufe 2
println("isDag(graph_i) = ", isDag(graph_i))
println("isDag(graph_ii) = ", isDag(graph_ii))