#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Starke Zusammenhangskomponenten (ZHKs)

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 75, 76, 77
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
include("./../Utils/stack.jl")

mutable struct ZhkData 
	oReps :: Stack 
	oNodes :: Stack
	dfsPos :: Int

	ZhkData() = new(make_stack(), make_stack(), 1)
end
mutable struct ZhkNode
	value :: Any
	marked :: Bool
	dfsNum :: Int
	component :: Union{Nothing, ZhkNode}

	ZhkNode(value) = new(value, false, 0, nothing)
end
function string(v::ZhkNode)
	return v.value
end

function zhk(s::ZhkNode, g::Graph)
	for v in nodes(g)
		v.marked = false
	end
	# init()
		zhkData = ZhkData()

	s.marked = true
	# root(s)
		push(s, zhkData.oReps) # neue ZHK
		push(s, zhkData.oNodes) # neuer offener Knoten
		s.dfsNum = zhkData.dfsPos
		zhkData.dfsPos += 1

	zhk(s, s, g, zhkData)
end
function zhk(u::ZhkNode, v::ZhkNode, g::Graph, zhkData::ZhkData)
	for (_, w) in edges_out(v, g)
		if w.marked  
			# handle_non_tree_edge(v, w)
				if w in zhkData.oNodes
					while w.dfsNum < top(zhkData.oReps).dfsNum
						pop(zhkData.oReps)
					end
				end

				if @isdefined(PRINT_STACKS) && PRINT_STACKS
					println("handle_non_tree_edge($(string(v)), $(string(w)))")
					println("\t oNodes: $(zhkData.oNodes)")
					println("\t oReps: $(zhkData.oReps)")
				end
		else
			# traverse_tree_dge(v,w)
				push(w, zhkData.oReps) # neue ZHK
				push(w, zhkData.oNodes) # neuer offener Knoten
				w.dfsNum = zhkData.dfsPos
				zhkData.dfsPos += 1

				if @isdefined(PRINT_STACKS) && PRINT_STACKS
					println("traverse_tree_edge($(string(v)), $(string(w)))")
					println("\t oNodes: $(zhkData.oNodes)")
					println("\t oReps: $(zhkData.oReps)")
				end

			w.marked = true 
			zhk(v, w, g,  zhkData)
		end
	end
	# backtrack(u, v)
		if v == top(zhkData.oReps)
			pop(zhkData.oReps)

			while true
				w = pop(zhkData.oNodes)
				w.component = v 
				if w == v
					break
				end
			end
		end

		if @isdefined(PRINT_STACKS) && PRINT_STACKS
			println("backtrack($(string(u)), $(string(v)))")
			println("\t oNodes: $(zhkData.oNodes)")
			println("\t oReps: $(zhkData.oReps)")
		end
end

function zhk_print(g::Graph)
	println(g)
	println("\tZHL results:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		println("\t\tdfsNum[$(u.value)] = $(u.dfsNum), component[$(u.value)] = $(u.component.value) ")
	end
end

# Beispielaufrufe
PRINT_STACKS = false
A, B, C = ZhkNode("A"), ZhkNode("B"), ZhkNode("C")
graph_i = DirectedGraph([B => C, C => A], "d-ZhkGraph-i")
zhk(B, graph_i)
zhk_print(graph_i)

# Graph anlog zu Folie 77
a, b, c, d, e = ZhkNode("a"), ZhkNode("b"), ZhkNode("c"), ZhkNode("d"), ZhkNode("e")
f, g, h, i = ZhkNode("f"), ZhkNode("g"), ZhkNode("h"), ZhkNode("i")
graph_ii = DirectedGraph([
	b => c, c => a, c => e, c => d, c => f, d => c, d => e,
	e => a, f => g, g => h, g => d, g => i, h => f, g => i
], "d-ZhkGraph-ii")
zhk(b, graph_ii)
zhk_print(graph_ii)