#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Kuerzeste Wege in DAGs (Directed Acyclic Graphs) mittels topologischer Sortierung der Knoten.
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 99
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Stack laden
include("./../Utils/stack.jl")


mutable struct DagNode
	dist::Float32
	parent::Union{Nothing, DagNode}
	visited::Bool
	value::Any
	DagNode(value) = new(Inf, nothing, false, value)
end
function string(v::DagNode)
	return v.value
end

"""
	Diese Funktion setzt voraus, dass der gegebene Graph g
	auch tatsaechlich ein gerichteter Graph ohne Kreise ist!
"""
function shortest_path_in_dag(s::DagNode, g::Graph)
	top_sort_stack = topological_sort(g)

	s.dist = 0 # s hat Distanz 0 zu sich selbst
	s.parent = s
	while !empty_stack(top_sort_stack)
		u = pop(top_sort_stack)
		for (u, v, weight) in edges_out(u, g)
			if v.dist > u.dist + weight
				v.dist = u.dist + weight
				v.parent = u
			end
		end
	end
end

function topological_sort(g::Graph)::Stack
	stack = make_stack()
	for v in nodes(g)
		if !v.visited
			topological_sort_helper(v, stack, g)
		end
	end
	return stack
end

function topological_sort_helper(v::DagNode, stack::Stack, g::Graph)::Nothing
	v.visited = true
	for (v, u) in edges_out(v, g)
		if !u.visited
			topological_sort_helper(u, stack, g)
		end
	end
	push(v, stack)
	nothing
end

function shortest_path_in_dag_print(g::Graph)
	println(g)
	println("\tDistance values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\t$(u.value): ")
		print("d[$(u.value)] = $(u.dist), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end


# Beispielaufrufe
a, b, c, d = DagNode("A"), DagNode("B"), DagNode("C"), DagNode("D")
e, f, g, h = DagNode("E"), DagNode("F"), DagNode("G"), DagNode("H")
i, j, k = DagNode("I"), DagNode("J"), DagNode("K")
g = DirectedWeightedGraph([
	a => b, a => c, b => d, c => g, c => h, d => e, d => f,
	e => g, e => i, f => i, g => h, h => i, i => j, i => k
], [
	1, 2, 1, 4, 3, 1, 2,
	1, 1, 2, 1, 2, 1, 2
], "e-DAG")
shortest_path_in_dag(a, g)
shortest_path_in_dag_print(g)