#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Dijkstras Algorithmus
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 100
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Prioritaetswarteschlange laden
include("./../Utils/binary-heap-pq.jl")

mutable struct DijkstraNode
	d::Float32
	parent::Union{Nothing, DijkstraNode}
	value::Any
	DijkstraNode(value) = new(Inf, nothing, value)
end
function string(v::DijkstraNode)
	return v.value
end
Base.hash(v::DijkstraNode) = hash(string(v))
Base.isequal(u::DijkstraNode, v::DijkstraNode) = isequal(string(u), string(v))

function dijkstra(s::DijkstraNode, g::Graph)
	s.d = 0
	s.parent = s
	q = build_pq([s], (x) -> x.d, (x, new_key) -> x.d = new_key)
	while !empty_pq(q)
		u = delete_min(q) # u: min. Distanz zu s in q
		for (u, v, cost) in edges_out(u, g)
			dv = v.d
			if isinf(dv) # v schon in q?
				v.d = u.d + cost
				v.parent = u
				insert(v, q)
			elseif dv > u.d + cost # aktualisiere v.d
				v.parent = u
				decrease_key(v, q, dv - (u.d + cost))
			end
		end
	end
end

function dijkstra_print(g::Graph)
	println(g)
	println("\tDijkstra values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\t$(u.value): ")
		print("d[$(u.value)] = $(lpad(u.d, 4, " ")), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end


# Beispielaufrufe
a, b, c, d, f = DijkstraNode("A"), DijkstraNode("B"), DijkstraNode("C"), DijkstraNode("D"), DijkstraNode("F")
g, h, i, j, k = DijkstraNode("G"), DijkstraNode("H"), DijkstraNode("I"), DijkstraNode("J"), DijkstraNode("K")
g = WeightedGraph([
	a => b, a => d, b => c, b => j, c => f, c => g, c => k,
	d => h, d => i, f => g, f => j, h => k, i => k
], [
	5, 2, 5, 2, 3, 4, 3, # Kantengewichte
	5, 1, 2, 1, 3, 10
], "f-DijkstraGraph")

dijkstra(a, g)
dijkstra_print(g)