#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Bellman-Ford Algorithmus
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 141
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")


mutable struct BfNode
	d::Float32
	parent::Union{Nothing, BfNode}
	value::Any
	BfNode(value) = new(Inf, nothing, value)
end
function string(v::BfNode)
	return v.value
end

function bellman_ford(s::BfNode, g::Graph)
	n = num_nodes(g)
	s.d = 0
	s.parent = s
	for i = 1:n-1 # aktualisiere Kosten fuer n-1 Runden
		for (v, w, cost) in edges(g)
			if w.d > v.d + cost # bessere Distanz möglich?
				w.d = v.d + cost
				w.parent = v
			end
		end
	end
	for (v, w, cost) in edges(g) # in n-ter Runde noch besser?
		if w.d > v.d + cost
			infect(w, g)
		end
	end
end

function infect(v::BfNode, g::Graph) # propagiere -Inf Kosten von v aus
	if v.d > -Inf # noch nicht markiert?
		v.d = -Inf
		for (v, w) in edges_out(v, g)
			infect(w, g)
		end
	end
end

function bellman_ford_print(g::Graph)
	println(g)
	println("\tBellman-Ford values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\td[$(u.value)] = $(lpad(u.d, 4, " ")), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end

# Beispielaufrufe
a, b, c, d, e = BfNode("A"), BfNode("B"), BfNode("C"), BfNode("D"), BfNode("E")
g = DirectedWeightedGraph([
	a => b, b => c, c => d, d => e, e => b
], [
	1, 1, -1, 0, -2 # Kantengewichte
], "h-BellmanFordGraph")

bellman_ford(a, g)
bellman_ford_print(g)