#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Kruskal-Algorithmus
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 197
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Union-Find Datenstruktur laden
include("../Utils/union-find.jl")


function kruskal(g::Graph)::Set
	T = Set()
	node_sets = Dict(v => make_set(v) for v in nodes(g)) # initialisiere einelem. Mengen fuer V
	S = sort(collect(edges(g)), by=(e) -> e[3]) # e = (u, v, cost), aufsteigend sortiert nach Kosten
	for (u, v, cost) in S
		if find(node_sets[u]) != find(node_sets[v]) # versch. Mengen
			push!(T, (u, v, cost))
			union(node_sets[u], node_sets[v]) # u und v in einer Menge
		end
	end
	return T
end

function tree_print(tree::Set)
	println("\tMinimaler Spannbaum:")
	for (u, v, cost) in sort(collect(tree), by=(e) -> e[3])
		println("\t\t$u - $v (Kantengewicht $cost)")
	end
	println("\tSumme der Kantengewichte: $(sum([cost for (u, v, cost) in tree]))")
end


# Beispielaufrufe
a, b, c, d = "A", "B", "C", "D"
e, f, g, h = "E", "F", "G", "H"
g = WeightedGraph([
	a => b, a => e, b => c, b => e, c => d, c => f,
	d => f, d => h, e => f, e => g, g => h
], [
	2, 1, 3, 2, 1, 3, # Kantengewichte
	5, 2, 2, 4, 3
], "i-KruskalGraph")

tree = kruskal(g)
println(g)
tree_print(tree)