#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Jarnik-Prim Algorithmus
		Fuer eine schnelle Implementierung sollte man *immer* die Julia-eigenen Algorithmen nutzen!

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 201
"""

# Datenstruktur laden 
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")
# Prioritaetswarteschlange laden
include("./../Utils/binary-heap-pq.jl")

mutable struct JpNode
	d::Float32
	parent::Union{Nothing, JpNode}
	finished::Bool
	value::Any
	JpNode(value) = new(Inf, nothing, false, value)
end
function string(v::JpNode)
	return v.value
end

function jarnik_prim(s::JpNode, g::Graph)
	s.d = 0
	s.parent = s # T anfangs nur aus s
	q = build_pq([s], (x) -> x.d, (x, new_key) -> x.d = new_key)
	while !empty_pq(q)
		u = delete_min(q) # u: min. Distanz zu T in q
		u.finished = true
		for (u, v, cost) in edges_out(u, g)
			if !v.finished && cost < v.d # aktualisiere v.d zu T
				v.parent = u
				if isinf(v.d) # v noch nicht in q?
					v.d = cost
					insert(v, q)
				else
					decrease_key(v, q, v.d - cost)
				end
			end
		end
	end
	return construct_tree(s, g)
end

function construct_tree(s::JpNode, g::Graph)::Set
	tree = Set()
	for v in nodes(g)
		if v != s
			push!(tree, (v, v.parent))
		end
	end
	return tree
end

function jarnik_prim_print(g::Graph)
	println(g)
	println("\tJarnik-Prim values:")
	nodes_sorted = sort(collect(nodes(g)), by=(x) -> (x.value))
	for u in nodes_sorted
		print("\t\td[$(u.value)] = $(lpad(u.d, 4, " ")), ")
		println("parent[$(u.value)] = $(isnothing(u.parent) ? "nothing" : u.parent.value)")
	end
end

function tree_print(tree::Set)
	println("\tMinimaler Spannbaum:")
	for (u, v) in sort(collect(tree), by=(e) -> e[1].value)
		println("\t\t$(u.value) - $(v.value)")
	end
end


# Beispielaufrufe
a, b, c, d, f = JpNode("A"), JpNode("B"), JpNode("C"), JpNode("D"), JpNode("F")
g, h, i, j, k = JpNode("G"), JpNode("H"), JpNode("I"), JpNode("J"), JpNode("K")
g = WeightedGraph([
	a => b, a => d, b => c, b => j, c => f, c => k,
	d => h, d => i, f => g, f => j, h => k, i => k
], [
	5, 2, 5, 2, 3, 3, # Kantengewichte
	5, 1, 4, 1, 3, 10
], "j-JarnikPrimGraph")

tree = jarnik_prim(a, g)
jarnik_prim_print(g)
tree_print(tree)