#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Euler
		Algorithmus 1 zur Prüfung auf Eulerweg/Eulerkreis in ungerichteten Graphen
		Laufzeit O(|V| * |E|)

	Foliensatz: 10-Graphen
	Ca. ab Folie(n): 243
"""

# Datenstruktur laden
USE_GRAPH_PLOTS = true 
include("./../Utils/graph.jl")

# Der Bezeichner fuer das leere Zeichen, global definiert
const BOT = "\u22A5"

function euler(g::Graph) :: String
	U = 0
	for v in nodes(g)
		if isodd(degree(v, g))
			U += 1
		end
	end
	if U == 0
		return "Eulerkreis"
	elseif U == 2
		return "Eulerweg"
	else
		return BOT
	end
end

function degree(v::Any, g::Graph) :: Int
	d = 0
	for (u,w) in edges(g)
		if u == v || w == v
			d += 1
		end
	end
	return d
end

# Beispielaufrufe
g_i = Graph([
		"a" => "b", "a" => "c", "a" => "d",
		"b" => "c", "b" => "d",
		"c" => "d", "c" => "e",
		"d" => "e"
	], "l-haus-vom-nikolaus"
)
println(g_i)
println("euler(g_i)")
result = euler(g_i)
println("Result: ", result)

g_ii = Graph([
		"a" => "b", "a" => "d",
		"b" => "c", "b" => "e", "b" => "f",
		"c" => "d",
		"d" => "e", "d" => "f"
	], "l-eulerkreis"
)
println(g_ii)
println("euler(g_ii)")
result = euler(g_ii)
println("Result: ", result)

g_iii = Graph([
		"a" => "b", "a" => "d", "a" => "e",
		"b" => "c", "b" => "d",
		"c" => "e",
		"d" => "e"
	], "l-nichts"
)
println(g_iii)
println("euler(g_iii)")
result = euler(g_iii)
println("Result: ", result)
