#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Laengste gemeinsame Teilsequenz (Longest Common Subsequence, LCS)

	Foliensatz: 12-Dynamische-Programmierung-Greedy-Verfahren
	Ca. ab Folie(n): 29
"""

# 0-indiziertes Array laden
# (Dies dient nur der verbesserten Modellierung, Index 0 kann dann einen leeren
# String repraesentieren. Alle Funktionen koennen genauso gut auf
# einem 1-indizierten Array arbeiten. )
include("./../Utils/zero-index-array.jl")

function lcs_length(X, Y)
	m = length(X)
	n = length(Y)
	c = ZeroIndexArray{Int}(undef, m+1, n+1)
	for i = 0:m
		c[i, 0] = 0
	end
	for j = 0:n
		c[0, j] = 0
	end
	for i = 1:m
		for j = 1:n
			if X[i] == Y[j]
				c[i, j] = c[i-1, j-1] + 1
			else
				c[i, j] = maximum([c[i-1, j], c[i, j-1]])
			end
		end
	end
	return c
end

function lcs_str(X, Y, m)
	lcs = ""
	i = size(m, 1)-1
	j = size(m, 2)-1
	while i > 0 && j > 0
		if m[i, j] == m[i-1, j-1] + 1
			lcs = X[i] * lcs
			i -= 1
			j -= 1
		elseif m[i, j] == m[i-1, j]
			i -= 1
		else
			j -= 1
		end
	end
	return lcs
end

function print_lcs_matrix(X, Y, m)
	pad_len = 4
	print(string(repeat(" ", pad_len+1), lpad("j", pad_len), " "))
	for j = 0:size(m, 2)-1
		print(string(lpad(j, pad_len), " "))
	end
	print("\n")
	print(string("i", repeat(" ", 2*pad_len+2),lpad("Y[j]", pad_len)))
	for j = 1:length(Y)
		print(string(lpad(Y[j], pad_len), " "))
	end
	print("\n")
	for i = 0:size(m, 1)-1
		print(string(i, " "))
		print(string(repeat(" ", pad_len), i == 0 ? lpad("X[i]", pad_len) : lpad(X[i], pad_len-1)))
		print(i == 0 ? "" : " ")
		for j = 0:size(m, 2)-1
			print(string(lpad(m[i, j], pad_len), " "))
		end
		print("\n")
	end
	print("\n")
	println("X = $X")
	println("Y = $Y")
	println("LCS: '$(lcs_str(X, Y, m))' mit der Laenge $(m[size(m, 1)-1, size(m, 2)-1])")
end

# Beispielaufrufe
X_1 = "ABCB"
Y_1 = "BDCAB"
print_lcs_matrix(X_1, Y_1, lcs_length(X_1, Y_1))

X_2 = "ACGTACC"
Y_2 = "AGTGCC"
print_lcs_matrix(X_2, Y_2, lcs_length(X_2, Y_2))