#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Rucksackproblem

	Foliensatz: 12-Dynamische-Programmierung-Greedy-Verfahren
	Ca. ab Folie(n): 61
"""

# 0-indiziertes Array laden
# (Dies dient nur der verbesserten Modellierung, Index 0 kann dann einen leeren
# String repraesentieren. Alle Funktionen koennen genauso gut auf
# einem 1-indizierten Array arbeiten. )
include("./../Utils/zero-index-array.jl")

function rucksack(items, W)
	n = length(items)
	V = ZeroIndexArray{Int}(undef, n+1, W+1)
	for i = 0:n
		V[i, 0] = 0
	end
	for w = 0:W
		V[0, w] = 0
	end
	for i = 1:n
		(wi, vi) = items[i]
		for w = 1:W
			if w - wi >= 0
				V[i, w] = maximum([V[i-1, w-wi] + vi, V[i-1, w]])
			else
				V[i, w] = V[i-1, w]
			end
		end
	end
	return V
end

function print_items(items, V)
	i = size(V, 1)-1
	w = size(V, 2)-1
	while i > 0 && w > 0
		(wi, vi) = items[i]
		if w - wi >= 0 && V[i, w] == V[i-1, w-wi] + vi
			println("Waehle Gegenstand $i mit Gewicht $wi und Wert $vi")
			i -= 1
			w -= wi
		else
			i -= 1
		end
	end
	println("Gesamtwert des Rucksacks: $(V[size(V, 1)-1, size(V, 2)-1])")
end

function print_matrix(items, V)
	pad_len = 4
	print(string(repeat(" ", pad_len+1), lpad("w", 3*pad_len+1), " "))
	for w = 0:size(V, 2)-1
		print(string(lpad(w, pad_len), " "))
	end
	print("\n")
	print(string("i", lpad("w_i", 2*pad_len+1), lpad("v_i", 2*pad_len+1)))
	for i = 0:size(V, 1)-1
		print(i != 0 ? string(i, " ") : "")
		print(i != 0 ? string(lpad(items[i][1], 2*pad_len-1), lpad(items[i][2], 2*pad_len+1)) : "")
		print(i != 0 ? " " : "")
		for w = 0:size(V, 2)-1
			print(string(lpad(V[i, w], pad_len), " "))
		end
		print("\n")
	end
	print("\n")
end

# Beispielaufrufe
items = [(2, 2), (4, 3), (3, 3), (5, 6), (2, 4), (6, 9)]
W = 10
V = rucksack(items, W)
print_matrix(items, V)
print_items(items, V)

items_ = [(1, 2), (2, 2), (2, 3), (4, 6), (5, 8), (7, 8)]
W_ = 8
V_ = rucksack(items_, W_)
print_matrix(items_, V_)
print_items(items_, V_)