#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Knuth-Morris-Pratt
		Algorithmus zum Suchen eines Musters in einem Text.

	Foliensatz: 13-Zeichenkettenabgleich
	Ca. ab Folie(n): 11
"""

function compute_f(p :: String) :: Array{Int}
	f = fill(1, length(p) - 1)
	m = length(f)
	i = 2
	j = 1
	while i <= m
		if p[j] == p[i] # j Zeichen stimmen überein
			f[i] = j + 1
			i += 1
			j += 1
		elseif j > 1 # j folgt dem übereinstimmenden Präfix
			j = f[j-1]
		else # keine Übereinstimmung
			f[i] = 1
			i += 1
		end
	end
	return f
end

function kmp_search(t :: String, p :: String) :: Int
	n = length(t)
	m = length(p)
	f = compute_f(p)
	i = 1
	j = 1
	while i <= n # passende Teilkette
		if p[j] == t[i]
			if j == m
				return i - m + 1 # erfolgreiche Suche
			end
			i += 1
			j += 1
		elseif j > 1 
			j = f[j-1]
		else
			i += 1
		end
	end
	return -1 # erfolglose Suche
end

# Beispielaufrufe
text = "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, \
		sed diam nonumy eirmod tempor invidunt ut labore et dolore magna \
		aliquyam erat, sed diam voluptua. At vero eos et accusam et justo \
		duo dolores et ea rebum."
pattern = "z"
println("kmp_search(\"", text, "\", \"", pattern, "\")")
position = kmp_search(text, pattern)
println(position > 0 ? "Muster gefunden an Stelle " * string(position) : "Muster nicht vorhanden", ".")
println("")
pattern = "ip"
println("kmp_search(\"", text, "\", \"", pattern, "\")")
position = kmp_search(text, pattern)
println(position > 0 ? "Muster gefunden an Stelle " * string(position) : "Muster nicht vorhanden", ".")
