#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Boyer-Moore
		Algorithmus zum Suchen eines Musters in einem Text. 

	Foliensatz: 13-Zeichenkettenabgleich
	Ca. ab Folie(n): 18
"""

function build_l(p :: String) :: Array{Int}
	l = fill(0, 128) # nur ASCII unterstuetzt
	for i = 1:length(p)
		l[Int(p[i])] = i
	end
	return l
end

function bm_search(t :: String, p :: String) :: Int
	l = build_l(p)
	n = length(t)
	m = length(p)
	i = m
	j = m
	while i <= n
		if p[j] == t[i]
			if j == 1
				return i # Suche erfolgreich
			else # Spiegeltechnik
				i -= 1
				j -= 1
			end
		else # Zeichensprungtechnik (last = last occurrence)
			last = l[Int(t[i])]
			i = i + m - min(j-1, last)
			j = m
		end
	end
	return -1 # Muster länger als Text oder Suche erfolglos
end

# Beispielaufrufe
text = "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, \
		sed diam nonumy eirmod tempor invidunt ut labore et dolore magna \
		aliquyam erat, sed diam voluptua. At vero eos et accusam et justo \
		duo dolores et ea rebum."
pattern = "z"
println("bm_search(\"", text, "\", \"", pattern, "\")")
position = bm_search(text, pattern)
println(position > 0 ? "Muster gefunden an Stelle " * string(position) : "Muster nicht vorhanden", ".")
println("")
pattern = "ip"
println("bm_search(\"", text, "\", \"", pattern, "\")")
position = bm_search(text, pattern)
println(position > 0 ? "Muster gefunden an Stelle " * string(position) : "Muster nicht vorhanden", ".")
