#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Malte Luttermann
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Prioritaetswarteschlange fuer die weitere Nutzung, z.B. in Dijkstras Algorithmus.

	Zum Anschauen der Implementierung bitte in "04-*/a-binaerer-heap.jl" gucken.

	> Hier bitte nichts aendern, damit der aufbauende Code keine Fehler verursacht!
"""

# Datentypen
abstract type PQ end
struct BinaererHeapPQ <: PQ
	internal_repr :: Array{Any}
	key :: Function
	set_key :: Function
	mapping :: Dict
end

# Funktionen extern
function build_pq(elements, key::Function, set_key::Function)::BinaererHeapPQ
	elements = convert(Vector{Any}, elements)

	mapping = Dict()
	for (i, e) in enumerate(elements)
		mapping[e] = i
	end

	pq = BinaererHeapPQ(elements, key, set_key, mapping)

	H = pq.internal_repr
	n = length(H)
	for i = parent(n):-1:1
		sift_down(i, pq)
	end
	return pq
end

function insert(e, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H) + 1

	pq.mapping[e] = n

	insert!(H, n, e)
	sift_up(n, pq)

	nothing
end

function min_element(pq::BinaererHeapPQ)
	return pq.internal_repr[1]
end 

function delete_min(pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)
	e = H[1]

	H[1] = H[n]
	pq.mapping[H[1]] = 1

	deleteat!(H, n)
	delete!(pq.mapping, e)

	sift_down(1, pq)

	return e
end

function Base.in(e::Any, pq::BinaererHeapPQ)
	return haskey(pq.mapping, e)
end

function empty_pq(pq::BinaererHeapPQ)
	return length(pq.internal_repr) == 0
end

function delete(e, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)

	i = get(pq.mapping, e, nothing)
	if !isnothing(i)
		H[i] = H[n]
		pq.mapping[H[i]] = i

		deleteat!(H, n)
		delete!(pq.mapping, e)

		sift_down(i, pq)
	end

	nothing
end

function decrease_key(e, pq::BinaererHeapPQ, delta)
	i = get(pq.mapping, e, nothing)
	if !isnothing(i)
		pq.set_key(e, pq.key(e) - delta)
		sift_up(i, pq)
	end 

	nothing
end

# Funktionen intern 
function sift_up(i::Int, pq::BinaererHeapPQ)
	H = pq.internal_repr

	while i > 1 && pq.key(H[parent(i)] ) > pq.key(H[i])
		temp = H[i]
		H[i] = H[parent(i)]
		H[parent(i)] = temp

		pq.mapping[H[i]] = i
		pq.mapping[H[parent(i)]] = parent(i)

		i = parent(i)
	end
end

function sift_down(i::Int, pq::BinaererHeapPQ)
	H = pq.internal_repr

	while !is_leaf(i, H)
		if !exists(right_child(i), H) 
			m = left_child(i)
		else 
			if pq.key(H[left_child(i)]) < pq.key(H[right_child(i)])
				m = left_child(i)
			else 
				m = right_child(i)
			end 
		end
		if pq.key(H[i]) <= pq.key(H[m])
			return # Invariante erfuellt
		else 
			temp = H[i]
			H[i] = H[m]
			H[m] = temp

			pq.mapping[H[i]] = i
			pq.mapping[H[m]] = m
		end
		i = m
	end
end

function parent(i)
    return i ÷ 2
end

function is_leaf(i, H)
    return 2 * i > length(H)
end

function exists(i, H)
    return i <= length(H)
end

function left_child(i)
    return 2 * i
end

function right_child(i)
    return 2 * i + 1
end

# Ausgabe als Baum durch Julia
function Base.show(io::IO, pq::BinaererHeapPQ)
	H = pq.internal_repr
	n = length(H)
	pad_size = 12

	println("BinaererHeapPQ mit ", n, " Elementen:")
	height = ceil(log2(n)) 
	width = Int(2^(height-1) * pad_size)
	for i = 1:height+1
		s = Int(2^(i-1))
		e = Int(2^i-1)
		pad = width ÷ (e - s + 1)
		for j = s:minimum([e, n])
			print(io, lpad(rpad(H[j], pad ÷ 2), pad))
		end
		println(io)
	end
end