#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 Henrik Strunck
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Circular Linked List fuer die weitere Nutzung z.B. in Binary Search Tree.

	> Hier bitte nichts aendern, damit der aufbauende Code keine Fehler verursacht!
"""

mutable struct CircularListElement
	value :: Any
	next :: Union{CircularListElement, Nothing}
	prev :: Union{CircularListElement, Nothing}
end

mutable struct CircularLinkedList
	start :: Union{CircularListElement, Nothing}
	last :: Union{CircularListElement, Nothing}
	length :: Int
end

function make_circular_list() :: CircularLinkedList
	return CircularLinkedList(nothing, nothing, 0)
end

function insert_first(e::Any, l::CircularLinkedList) :: CircularListElement
	old_start = l.start
	l.start = CircularListElement(e, old_start, nothing)
	if length(l) == 0
		l.last = l.start
		l.start.next = l.start
	else
		l.last.next = l.start
		old_start.prev = l.start
	end
	l.start.prev = l.last
	l.length += 1
	return l.start
end

function insert_last(e::Any, l::CircularLinkedList) :: CircularListElement
	old_last = l.last
	l.last = CircularListElement(e, nothing, old_last)
	if length(l) == 0
		l.start = l.last
	else
		old_last.next = l.last
	end
	l.last.next = l.start
	l.start.prev = l.last
	l.length += 1
	return l.last
end

function insert_before(e::Any, le::CircularListElement, l::CircularLinkedList) :: CircularListElement
	new_element = CircularListElement(e, le, nothing)
	if le == l.start # insert in front of first element
		l.start = new_element
		new_element.prev = l.last
		l.last.next = new_element
	else
		le.prev.next = new_element
		new_element.prev = le.prev
	end
	le.prev = new_element
	l.length += 1
	return new_element
end

function delete(e::Any, l::CircularLinkedList) :: Nothing
	c = l.start
	while !isnothing(c)
		if c.value == e
			delete(c, l)
			break
		end
		c = c.next
	end
	nothing
end
function delete(le::CircularListElement, l::CircularLinkedList) :: Nothing
	if le == l.start # erstes Element (Kopf anpassen)
		le.next.prev = l.last
		l.last.next = le.next
		l.start = le.next
	else
		le.prev.next = le.next
		le.next.prev = le.prev
	end
	if l.last == le # letztes Element geloescht?
		l.last = le.prev
	end
	l.length -= 1
	nothing
end

function first(l::CircularLinkedList) :: Any
	if isnothing(l.start)
		throw(BoundsError(l, 1))
	else 
		return l.start.value
	end
end

function delete_first(l::CircularLinkedList) :: Nothing
	if isnothing(l.start)
		throw(BoundsError(l, 1))
	else
		delete(l.start.value, l)
	end
	nothing
end

function empty_list(l::CircularLinkedList) :: Bool
	return l.length == 0
end

function concat(l::CircularLinkedList, l_::CircularLinkedList) :: Nothing
	if empty_list(l)
		l.start = l_.start 
	elseif !empty_list(l_)
		l.start.prev = l_.last
		l.last.next = l_.start
	end
	if !empty_list(l_)
		l_.start.prev = l.last
		l_.last.next = l.start
		l.last = l_.last
	end
	l.length += l_.length
	nothing
end

function Base.iterate(l::CircularLinkedList) :: Union{Nothing, Tuple{Any, Union{CircularListElement, Nothing}}}
	if isnothing(l.start)
		return nothing
	elseif l.start.next == l.start
		return l.start.value, nothing
	else
		return l.start.value, l.start.next
	end 
end
function Base.iterate(l::CircularLinkedList, state::CircularListElement) :: Tuple{Any, Union{CircularListElement, Nothing}}
	if state.next == l.start
		return state.value, nothing
	else
		return state.value, state.next
	end
end
function Base.iterate(l::CircularLinkedList, state::Nothing) :: Nothing
	return nothing
end
function Base.length(l::CircularLinkedList) :: Int
	return l.length
end

function Base.show(io::IO, l::CircularLinkedList)
	println(io, "Circular Linked List mit $(length(l)) Elementen:")
	if !empty_list(l)
		print(io, "\t$(l.start.value) -> [ $(join(map(e->e, l), " <-> ")) ] <- $(l.last.value)")
	else 
		print("\tnothing -> [ ] <- nothing")
	end
end
