
#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Simple wrapper to filter and process iterable elements
	in a generator like way.

		Type `Generator` is iterable and yields values from given 
		iterable `iter`. But before yielding a value, the values 
		are processed by `process_function`, e.g., only yield parts of 
		elements in iter, and all elements are filtered by `filter_function` 
		and only elements where `filter_function` returns true are 
		returned by the iterators.

	> Please do not change the code here, otherwise some examples using this
	> code may fail!

	*It is not necessary to understand this code.*
"""

mutable struct Generator
	process_function :: Function
	iter
	filter_function :: Function
	length :: Int

	Generator(p_f::Function, iter) = new(p_f, iter, (e) -> true, -1)
	Generator(iter, f_f::Function) = new(identity, iter, f_f, -1)
	Generator(p_f::Function, iter, f_f::Function) = new(p_f, iter, f_f, -1)
end

function Base.length(g::Generator)
	if g.length == -1
			len = 0
			for _ in g
				len += 1
			end
			g.length = len
	end
	return g.length
end

function Base.iterate(g::Generator, state=nothing) 
	if isnothing(state)
		element_state = iterate(g.iter) # start iterate
	else 
		element_state = iterate(g.iter, state) # continue iterate
	end

	while !isnothing(element_state)
		(element, state) = element_state

		if g.filter_function(element)
			return g.process_function(element), state 
		end
		element_state = iterate(g.iter, state)
	end 

	return nothing  
end 