#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Diese Datei spezifiziert ein einfaches Mapping von (positiven) Integers auf 
	beliebige Werte/ Objekte. 

	**Hinweis**
	Dieses Mapping kann man in PQs nutzen, um Elemente anhand ihres Key 
	zu identifizieren. So kann man z.B. `delete(Person("x"), PQ)` ausfuehren,
	auch wenn die PQ selbst intern andere Objekte nutzt, die nur einen Verweis auf Person("x")
	beinhalten. Aus dem Aufruf wird dann quasi `delete(lookup(key(Person("x"))), PQ)``
	und `lookup()` gibt ein intern von PQ genutzes Objekt zurueck, welches dann 
	geloescht werden kann.

	Das Mapping hier hat einige Nachteile, da man einfach ein Array in der 
	Laenge des maximalen Key erzeugt und dann im Array unter dem Index des Key 
	das Objekt ablegt (keine negativen Keys, grosses duenn besetztes Array, ...).

	In der Vorlesung werden wir noch Hashing kennen lernen, damit kann man dann ein 
	Mapping wie hier dargestellt ohne riesige Arrays und fuer beliebige Keys implementieren.

	> Hier bitte nichts aendern, damit der aufbauende Code keine Fehler verursacht!
"""

struct KeyMapping
	key_element :: Array{Any} # key(element) => element
end

function KeyMapping(max_key=10) :: KeyMapping
	return KeyMapping(Array{Any}(nothing, max_key))
end

function lookup(k::Int, m::KeyMapping) :: Any
	if k > 0 && k <= length(m.key_element) 
		return m.key_element[k]
	else
		return nothing
	end
end

function insert(k::Int, e::Any, m::KeyMapping) :: Nothing
	if k > 0 
		if k > length(m.key_element) # Array vergroessern
			append!(m.key_element, fill(nothing, k - length(m.key_element)))
		end 

		if isnothing(m.key_element[k]) # fuege ein, falls nicht vergeben
			m.key_element[k] = e
		else 
			throw(ArgumentError("Keys have to be distinct! ($e, key=$k)"))
		end
	else
		throw(ArgumentError("Only positive keys are allowed! ($e, key=$k)"))
	end

	nothing
end

function update(k::Int, e::Any, m::KeyMapping) :: Nothing
	if k > 0 && k <= length(m.key_element)
		m.key_element[k] = e
	else 
		throw(ArgumentError("Unknown key to update! ($e, key=$k)"))
	end

	nothing
end

function delete(k::Int, m::KeyMapping) :: Nothing
	if k > 0 && k <= length(m.key_element) 
		m.key_element[k] = nothing
	end

	nothing
end

function update_key(old_key::Int, new_key::Int, m::KeyMapping) :: Nothing
	e = lookup(old_key, m)
	if !isnothing(e)
		delete(old_key, m)
		insert(new_key, e, m)
	end

	nothing
end

function merge(m::KeyMapping, m_::KeyMapping, f::Function=identity)
	if m != m_ 
		for (key, value) in enumerate(m_.key_element)
			if !isnothing(value)
				insert(key, f(value), m)
			end
		end
	end
end

function Base.show(io::IO, m::KeyMapping)
	str = "\t"
	cnt = 1
	for (key, value) in enumerate(m.key_element)
		if !isnothing(value)
			str *= string(key, " => ", value, ", ")
			if cnt % 3 == 0
				str *= "\n\t"
			end
			cnt += 1
		end
	end

	print(io, "[\n" * str[begin:end-2] * "\n]") 
end