#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Linked List fuer die weitere Nutzung z.B. in Binomial Heaps.

	Zum Anschauen der Implementierung bitte in "03-*/c-linked-list.jl" gucken.

	> Hier bitte nichts aendern, damit der aufbauende Code keine Fehler verursacht!
"""

mutable struct ListElement
	value :: Any
	next :: Union{ListElement, Nothing}
end

mutable struct LinkedList 
	start :: Union{ListElement, Nothing}
	last :: Union{ListElement, Nothing}
	length :: Int
end

function make_list() :: LinkedList
	return LinkedList(nothing, nothing, 0)
end

function insert(e::Any, l::LinkedList) :: Nothing
	l.start = ListElement(e, l.start)
	if length(l) == 0
		l.last = l.start 
	end
	l.length += 1
	nothing
end

function cons(e::Any, l::LinkedList) :: LinkedList
	l_ = LinkedList(
		ListElement(e, l.start),
		l.last,
		l.length+1
	)
	if length(l) == 0
		l_.last = l_.start
	end
	return l_
end

function delete(e::Any, l::LinkedList) :: Nothing
	c = l.start 
	b = nothing
	while !isnothing(c)
		if c.value == e 
			if isnothing(b) # erstes Element (Kopf anpassen)
				l.start = c.next 
			else
				b.next = c.next
			end
			if l.last.value == e # letztes Element geloescht?
				l.last = b
			end
			l.length -= 1
			break
		end
		b = c
		c = c.next
	end
	nothing
end

function first(l::LinkedList) :: Any
	if isnothing(l.start)
		throw(BoundsError(l, 1))
	else 
		return l.start.value
	end
end

function delete_first(l::LinkedList) :: Nothing
	if isnothing(l.start)
		throw(BoundsError(l, 1))
	else
		delete(l.start.value, l)
	end
	nothing
end

function empty_list(l::LinkedList) :: Bool
	return l.length == 0
end

function concat(l::LinkedList, l_::LinkedList) :: Nothing
	if empty_list(l)
		l.start = l_.start 
	else
		l.last.next = l_.start 
	end

	if !empty_list(l_)
		l.last = l_.last
	end

	l.length += l_.length
	nothing
end

function Base.iterate(l::LinkedList) :: Union{Nothing, Tuple{Any, Union{ListElement, Nothing}}}
	if isnothing(l.start)
		return nothing 
	else 
		return l.start.value, l.start.next
	end 
end 
function Base.iterate(l::LinkedList, state::ListElement) :: Tuple{Any, Union{ListElement, Nothing}}
	return state.value, state.next 
end
function Base.iterate(l::LinkedList, state::Nothing) :: Nothing
	return nothing
end 
function Base.length(l::LinkedList) :: Int
	return l.length
end

function Base.show(io::IO, l::LinkedList)
	println(io, "Linked List mit $(length(l)) Elementen:")
	if !empty_list(l)
		print(io, "  -> $(join(map(e->e, l), " -> ")); $(l.last.value) <- ")
	end
end