#=
 Vorlesung
	Algorithmen und Datenstrukturen
	Prof. Dr. Ralf Moeller
	Institut fuer Informationssysteme, Universitaet zu Luebeck
	https://www.ifis.uni-luebeck.de

 Magnus Bender
 (c) 2022 - 2023, GPLv3 gnu.org/licenses/gpl-3.0.txt
=#

"""
	Union Find mit Pfadkompression fuer die weitere Nutzung, z.B. bei minimalen Spannbaeumen.

	Zum Anschauen der Implementierung bitte in "08-*/a-union-find.jl" gucken.

	> Hier bitte nichts aendern, damit der aufbauende Code keine Fehler verursacht!
"""

# Linked List analog zu Foliensatz 03, Folie 54 laden
include("./linked-list.jl")

mutable struct UnionFindElement 
	value :: Any
	parent :: Union{Nothing, UnionFindElement}
	rank :: Int
		# Nach find() mit Pfadkompression ist der Rang nur noch
		# eine obere Schranke, da Updates am Rang nur bei union() stattfinden,
		# siehe: Robert E. Tarjan, Jan van Leeuwen. Worst-case analysis of
		# set union algorithms, Journal of the ACM 31 (2), S. 245–281, 1984
end 

# Zugriff auf Attribute
function value(x::UnionFindElement) :: Any
	return x.value
end

function rank(x::UnionFindElement) :: Int
	return x.rank
end

# Operationen 
function make_set(e::Any) :: UnionFindElement
	return UnionFindElement(e, nothing, 0)
end

function find(x::UnionFindElement) :: UnionFindElement
	visited = make_list()
	while !isnothing(x.parent)
		insert(x, visited)
		x = x.parent
	end
	representant = x

	for x_ in visited # Pfadkompression 
		x_.parent = representant 
	end
	
	return representant
end

function union(x::UnionFindElement, x_::UnionFindElement) :: Nothing
	x = find(x)
	x_ = find(x_)

	if x != x_
		if rank(x) >= rank(x_)
			x_.parent = x 
		else 
			x.parent = x_
		end

		if rank(x) == rank(x_)
			x.rank += 1
		end
	end

	nothing
end

# Ausgabe 
function Base.show(io::IO, x::UnionFindElement) 
	str = ""
	while !isnothing(x)
		str *= string(value(x), " (", x.rank ,") -> ")
		x = x.parent
	end
	print(io, str[begin:end-4])
end
function Base.show(io::IO, elems::Vector{UnionFindElement}) 
	println(io, "Union Find Elemente [Element (Rang) -> ... -> Repraesentant]:")
	for x in elems
		println(io, "\t", x)
	end
end